
package org.mule.module.twilio.config;

import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.lang.reflect.WildcardType;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;
import org.mule.api.MessagingException;
import org.mule.api.MuleContext;
import org.mule.api.MuleEvent;
import org.mule.api.MuleException;
import org.mule.api.MuleMessage;
import org.mule.api.callback.HttpCallback;
import org.mule.api.construct.FlowConstruct;
import org.mule.api.construct.FlowConstructAware;
import org.mule.api.context.MuleContextAware;
import org.mule.api.expression.ExpressionManager;
import org.mule.api.lifecycle.Disposable;
import org.mule.api.lifecycle.Initialisable;
import org.mule.api.lifecycle.InitialisationException;
import org.mule.api.lifecycle.Startable;
import org.mule.api.lifecycle.Stoppable;
import org.mule.api.processor.MessageProcessor;
import org.mule.api.registry.RegistrationException;
import org.mule.api.transformer.DataType;
import org.mule.api.transformer.Transformer;
import org.mule.api.transformer.TransformerException;
import org.mule.config.i18n.CoreMessages;
import org.mule.config.i18n.MessageFactory;
import org.mule.construct.Flow;
import org.mule.module.twilio.TwilioConnector;
import org.mule.module.twilio.config.spring.DefaultHttpCallback;
import org.mule.transformer.TransformerTemplate;
import org.mule.transformer.types.DataTypeFactory;
import org.mule.transport.NullPayload;
import org.mule.util.TemplateParser;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


/**
 * MakeCallMessageProcessor invokes the {@link org.mule.module.twilio.TwilioConnector#makeCall(java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String, org.mule.module.twilio.HttpMethod, org.mule.api.callback.HttpCallback, org.mule.api.callback.HttpCallback, java.lang.String, java.lang.String, java.lang.String)} method in {@link TwilioConnector }. For each argument there is a field in this processor to match it.  Before invoking the actual method the processor will evaluate and transform where possible to the expected argument type.
 * 
 */
public class MakeCallMessageProcessor
    implements FlowConstructAware, MuleContextAware, Disposable, Initialisable, Startable, Stoppable, MessageProcessor
{

    private Object accountSid;
    private String _accountSidType;
    private Object from;
    private String _fromType;
    private Object to;
    private String _toType;
    private Object url;
    private String _urlType;
    private Object applicationSid;
    private String _applicationSidType;
    private Object method;
    private org.mule.module.twilio.HttpMethod _methodType;
    /**
     * The flow to be invoked when the http callback is received
     * 
     */
    private Flow fallbackCallbackFlow;
    /**
     * An HttpCallback instance responsible for linking the APIs http callback with the flow {@link org.mule.module.twilio.config.MakeCallMessageProcessor#fallbackCallbackFlow
     * 
     */
    private HttpCallback fallback;
    /**
     * The flow to be invoked when the http callback is received
     * 
     */
    private Flow statusCallbackCallbackFlow;
    /**
     * An HttpCallback instance responsible for linking the APIs http callback with the flow {@link org.mule.module.twilio.config.MakeCallMessageProcessor#statusCallbackCallbackFlow
     * 
     */
    private HttpCallback statusCallback;
    private Object sendDigits;
    private String _sendDigitsType;
    private Object ifMachine;
    private String _ifMachineType;
    private Object timeout;
    private String _timeoutType;
    private static Logger logger = LoggerFactory.getLogger(MakeCallMessageProcessor.class);
    /**
     * Module object
     * 
     */
    private Object moduleObject;
    /**
     * Mule Context
     * 
     */
    private MuleContext muleContext;
    /**
     * Mule Expression Manager
     * 
     */
    private ExpressionManager expressionManager;
    /**
     * Mule Pattern Info
     * 
     */
    private TemplateParser.PatternInfo patternInfo;
    /**
     * Flow construct
     * 
     */
    private FlowConstruct flowConstruct;
    /**
     * Variable used to track how many retries we have attempted on this message processor
     * 
     */
    private AtomicInteger retryCount;
    /**
     * Maximum number of retries that can be attempted.
     * 
     */
    private int retryMax;

    /**
     * Obtains the expression manager from the Mule context and initialises the connector. If a target object  has not been set already it will search the Mule registry for a default one.
     * 
     * @throws InitialisationException
     */
    public void initialise()
        throws InitialisationException
    {
        retryCount = new AtomicInteger();
        expressionManager = muleContext.getExpressionManager();
        patternInfo = TemplateParser.createMuleStyleParser().getStyle();
        if (moduleObject == null) {
            try {
                moduleObject = muleContext.getRegistry().lookupObject(TwilioConnectorHttpCallbackAdapter.class);
                if (moduleObject == null) {
                    throw new InitialisationException(MessageFactory.createStaticMessage("Cannot find object"), this);
                }
            } catch (RegistrationException e) {
                throw new InitialisationException(CoreMessages.initialisationFailure("org.mule.module.twilio.config.TwilioConnectorHttpCallbackAdapter"), e, this);
            }
        }
        if (statusCallbackCallbackFlow!= null) {
            statusCallback = new DefaultHttpCallback(statusCallbackCallbackFlow, muleContext, ((TwilioConnectorHttpCallbackAdapter) moduleObject).getDomain(), ((TwilioConnectorHttpCallbackAdapter) moduleObject).getLocalPort(), ((TwilioConnectorHttpCallbackAdapter) moduleObject).getRemotePort(), ((TwilioConnectorHttpCallbackAdapter) moduleObject).getAsync());
        }
        if (fallbackCallbackFlow!= null) {
            fallback = new DefaultHttpCallback(fallbackCallbackFlow, muleContext, ((TwilioConnectorHttpCallbackAdapter) moduleObject).getDomain(), ((TwilioConnectorHttpCallbackAdapter) moduleObject).getLocalPort(), ((TwilioConnectorHttpCallbackAdapter) moduleObject).getRemotePort(), ((TwilioConnectorHttpCallbackAdapter) moduleObject).getAsync());
        }
    }

    public void start()
        throws MuleException
    {
        if (statusCallback!= null) {
            statusCallback.start();
        }
        if (fallback!= null) {
            fallback.start();
        }
    }

    public void stop()
        throws MuleException
    {
        if (statusCallback!= null) {
            statusCallback.stop();
        }
        if (fallback!= null) {
            fallback.stop();
        }
    }

    public void dispose() {
    }

    /**
     * Set the Mule context
     * 
     * @param context Mule context to set
     */
    public void setMuleContext(MuleContext context) {
        this.muleContext = context;
    }

    /**
     * Sets flow construct
     * 
     * @param flowConstruct Flow construct to set
     */
    public void setFlowConstruct(FlowConstruct flowConstruct) {
        this.flowConstruct = flowConstruct;
    }

    /**
     * Sets the instance of the object under which the processor will execute
     * 
     * @param moduleObject Instace of the module
     */
    public void setModuleObject(Object moduleObject) {
        this.moduleObject = moduleObject;
    }

    /**
     * Sets retryMax
     * 
     * @param value Value to set
     */
    public void setRetryMax(int value) {
        this.retryMax = value;
    }

    /**
     * Sets to
     * 
     * @param value Value to set
     */
    public void setTo(Object value) {
        this.to = value;
    }

    /**
     * Sets statusCallbackCallbackFlow
     * 
     * @param value Value to set
     */
    public void setStatusCallbackCallbackFlow(Flow value) {
        this.statusCallbackCallbackFlow = value;
    }

    /**
     * Sets fallbackCallbackFlow
     * 
     * @param value Value to set
     */
    public void setFallbackCallbackFlow(Flow value) {
        this.fallbackCallbackFlow = value;
    }

    /**
     * Sets accountSid
     * 
     * @param value Value to set
     */
    public void setAccountSid(Object value) {
        this.accountSid = value;
    }

    /**
     * Sets ifMachine
     * 
     * @param value Value to set
     */
    public void setIfMachine(Object value) {
        this.ifMachine = value;
    }

    /**
     * Sets applicationSid
     * 
     * @param value Value to set
     */
    public void setApplicationSid(Object value) {
        this.applicationSid = value;
    }

    /**
     * Sets method
     * 
     * @param value Value to set
     */
    public void setMethod(Object value) {
        this.method = value;
    }

    /**
     * Sets sendDigits
     * 
     * @param value Value to set
     */
    public void setSendDigits(Object value) {
        this.sendDigits = value;
    }

    /**
     * Sets from
     * 
     * @param value Value to set
     */
    public void setFrom(Object value) {
        this.from = value;
    }

    /**
     * Sets timeout
     * 
     * @param value Value to set
     */
    public void setTimeout(Object value) {
        this.timeout = value;
    }

    /**
     * Sets url
     * 
     * @param value Value to set
     */
    public void setUrl(Object value) {
        this.url = value;
    }

    /**
     * Get all superclasses and interfaces recursively.
     * 
     * @param classes List of classes to which to add all found super classes and interfaces.
     * @param clazz   The class to start the search with.
     */
    private void computeClassHierarchy(Class clazz, List classes) {
        for (Class current = clazz; (current!= null); current = current.getSuperclass()) {
            if (classes.contains(current)) {
                return ;
            }
            classes.add(current);
            for (Class currentInterface: current.getInterfaces()) {
                computeClassHierarchy(currentInterface, classes);
            }
        }
    }

    /**
     * Checks whether the specified class parameter is an instance of {@link List }
     * 
     * @param clazz <code>Class</code> to check.
     * @return
     */
    private boolean isListClass(Class clazz) {
        List<Class> classes = new ArrayList<Class>();
        computeClassHierarchy(clazz, classes);
        return classes.contains(List.class);
    }

    /**
     * Checks whether the specified class parameter is an instance of {@link Map }
     * 
     * @param clazz <code>Class</code> to check.
     * @return
     */
    private boolean isMapClass(Class clazz) {
        List<Class> classes = new ArrayList<Class>();
        computeClassHierarchy(clazz, classes);
        return classes.contains(Map.class);
    }

    private boolean isList(Type type) {
        if ((type instanceof Class)&&isListClass(((Class) type))) {
            return true;
        }
        if (type instanceof ParameterizedType) {
            return isList(((ParameterizedType) type).getRawType());
        }
        if (type instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) type).getUpperBounds();
            return ((upperBounds.length!= 0)&&isList(upperBounds[ 0 ]));
        }
        return false;
    }

    private boolean isMap(Type type) {
        if ((type instanceof Class)&&isMapClass(((Class) type))) {
            return true;
        }
        if (type instanceof ParameterizedType) {
            return isMap(((ParameterizedType) type).getRawType());
        }
        if (type instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) type).getUpperBounds();
            return ((upperBounds.length!= 0)&&isMap(upperBounds[ 0 ]));
        }
        return false;
    }

    private boolean isAssignableFrom(Type expectedType, Class clazz) {
        if (expectedType instanceof Class) {
            return ((Class) expectedType).isAssignableFrom(clazz);
        }
        if (expectedType instanceof ParameterizedType) {
            return isAssignableFrom(((ParameterizedType) expectedType).getRawType(), clazz);
        }
        if (expectedType instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) expectedType).getUpperBounds();
            if (upperBounds.length!= 0) {
                return isAssignableFrom(upperBounds[ 0 ], clazz);
            }
        }
        return false;
    }

    private Object evaluate(MuleMessage muleMessage, Object source) {
        if (source instanceof String) {
            String stringSource = ((String) source);
            if (stringSource.startsWith(patternInfo.getPrefix())&&stringSource.endsWith(patternInfo.getSuffix())) {
                return expressionManager.evaluate(stringSource, muleMessage);
            } else {
                return expressionManager.parse(stringSource, muleMessage);
            }
        }
        return source;
    }

    private Object evaluateAndTransform(MuleMessage muleMessage, Type expectedType, Object source)
        throws TransformerException
    {
        if (source == null) {
            return source;
        }
        Object target = null;
        if (isList(source.getClass())) {
            if (isList(expectedType)) {
                List newList = new ArrayList();
                Type valueType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 0 ];
                ListIterator iterator = ((List) source).listIterator();
                while (iterator.hasNext()) {
                    Object subTarget = iterator.next();
                    newList.add(evaluateAndTransform(muleMessage, valueType, subTarget));
                }
                target = newList;
            } else {
                target = source;
            }
        } else {
            if (isMap(source.getClass())) {
                if (isMap(expectedType)) {
                    Type keyType = Object.class;
                    Type valueType = Object.class;
                    if (expectedType instanceof ParameterizedType) {
                        keyType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 0 ];
                        valueType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 1 ];
                    }
                    Map map = ((Map) source);
                    Map newMap = new HashMap();
                    for (Object entryObj: map.entrySet()) {
                        {
                            Map.Entry entry = ((Map.Entry) entryObj);
                            Object newKey = evaluateAndTransform(muleMessage, keyType, entry.getKey());
                            Object newValue = evaluateAndTransform(muleMessage, valueType, entry.getValue());
                            newMap.put(newKey, newValue);
                        }
                    }
                    target = newMap;
                } else {
                    target = source;
                }
            } else {
                target = evaluate(muleMessage, source);
            }
        }
        if ((target!= null)&&(!isAssignableFrom(expectedType, target.getClass()))) {
            DataType sourceDataType = DataTypeFactory.create(target.getClass());
            DataType targetDataType = DataTypeFactory.create(((Class) expectedType));
            Transformer t = muleContext.getRegistry().lookupTransformer(sourceDataType, targetDataType);
            return t.transform(target);
        } else {
            return target;
        }
    }

    /**
     * Invokes the MessageProcessor.
     * 
     * @param event MuleEvent to be processed
     * @throws MuleException
     */
    public MuleEvent process(MuleEvent event)
        throws MuleException
    {
        MuleMessage muleMessage = event.getMessage();
        TwilioConnectorHttpCallbackAdapter castedModuleObject = null;
        if (moduleObject instanceof String) {
            castedModuleObject = ((TwilioConnectorHttpCallbackAdapter) muleContext.getRegistry().lookupObject(((String) moduleObject)));
            if (castedModuleObject == null) {
                throw new MessagingException(CoreMessages.failedToCreate("makeCall"), event, new RuntimeException("Cannot find the configuration specified by the config-ref attribute."));
            }
        } else {
            castedModuleObject = ((TwilioConnectorHttpCallbackAdapter) moduleObject);
        }
        try {
            String transformedAccountSid = ((String) evaluateAndTransform(muleMessage, MakeCallMessageProcessor.class.getDeclaredField("_accountSidType").getGenericType(), accountSid));
            String transformedFrom = ((String) evaluateAndTransform(muleMessage, MakeCallMessageProcessor.class.getDeclaredField("_fromType").getGenericType(), from));
            String transformedTo = ((String) evaluateAndTransform(muleMessage, MakeCallMessageProcessor.class.getDeclaredField("_toType").getGenericType(), to));
            String transformedUrl = ((String) evaluateAndTransform(muleMessage, MakeCallMessageProcessor.class.getDeclaredField("_urlType").getGenericType(), url));
            String transformedApplicationSid = ((String) evaluateAndTransform(muleMessage, MakeCallMessageProcessor.class.getDeclaredField("_applicationSidType").getGenericType(), applicationSid));
            org.mule.module.twilio.HttpMethod transformedMethod = ((org.mule.module.twilio.HttpMethod) evaluateAndTransform(muleMessage, MakeCallMessageProcessor.class.getDeclaredField("_methodType").getGenericType(), method));
            String transformedSendDigits = ((String) evaluateAndTransform(muleMessage, MakeCallMessageProcessor.class.getDeclaredField("_sendDigitsType").getGenericType(), sendDigits));
            String transformedIfMachine = ((String) evaluateAndTransform(muleMessage, MakeCallMessageProcessor.class.getDeclaredField("_ifMachineType").getGenericType(), ifMachine));
            String transformedTimeout = ((String) evaluateAndTransform(muleMessage, MakeCallMessageProcessor.class.getDeclaredField("_timeoutType").getGenericType(), timeout));
            retryCount.getAndIncrement();
            Object resultPayload;
            resultPayload = castedModuleObject.makeCall(transformedAccountSid, transformedFrom, transformedTo, transformedUrl, transformedApplicationSid, transformedMethod, fallback, statusCallback, transformedSendDigits, transformedIfMachine, transformedTimeout);
            TransformerTemplate.OverwitePayloadCallback overwritePayloadCallback = null;
            if (resultPayload == null) {
                overwritePayloadCallback = new TransformerTemplate.OverwitePayloadCallback(NullPayload.getInstance());
            } else {
                overwritePayloadCallback = new TransformerTemplate.OverwitePayloadCallback(resultPayload);
            }
            List<Transformer> transformerList;
            transformerList = new ArrayList<Transformer>();
            transformerList.add(new TransformerTemplate(overwritePayloadCallback));
            event.getMessage().applyTransformers(event, transformerList);
            retryCount.set(0);
            return event;
        } catch (Exception e) {
            throw new MessagingException(CoreMessages.failedToInvoke("makeCall"), event, e);
        }
    }

}
